import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:sunny_core_widgets/routes/router_factory.dart';
import 'package:sunny_core_widgets/sunny_core_widgets.dart';
import 'package:sunny_fluro/sunny_fluro.dart';
import 'package:sunny_lifecycle/routing/route_history.dart';

class TestCaptureHistoryObserver implements HistoryObserver {
  final List<HistoryEvent> events = [];

  @override
  void didPopHistory(RouteHistoryItem item, List<RouteHistoryItem> snapshot) {
    events.insert(
      0,
      HistoryEvent(
        didPush: false,
        item: item,
        snapshot: snapshot,
      ),
    );
  }

  @override
  void didPushHistory(RouteHistoryItem item, List<RouteHistoryItem> snapshot) {
    events.insert(
      0,
      HistoryEvent(
        didPush: true,
        item: item,
        snapshot: snapshot,
      ),
    );
  }
}

class HistoryEvent {
  final bool didPush;
  final RouteHistoryItem item;
  final List<RouteHistoryItem> snapshot;

  HistoryEvent(
      {required this.didPush, required this.item, required this.snapshot});

  @override
  String toString() {
    return '${didPush ? 'Pushed' : 'Popped'} ${item.route} - ROUTES: ${snapshot.map((ev) => ev.route).join(' => ')}';
  }
}

typedef WidgetListBuilder = List<Widget> Function(BuildContext context);

FRouter newTestRouter() {
  return FRouter.ofUriTemplates(
    routeFactory: const SunnyRouterFactory(),
    notFoundRoute: NotFoundRoute,

    /// While we're using [Get], pass the global nav key
    navigatorOf: (BuildContext context, isRoot) =>
        Navigator.of(context, rootNavigator: isRoot),
  );
}

extension WidgetTesterExt on WidgetTester {
  Future changeTab(int num) async {
    var pageSwitcher =
        find.widgetWithText(MaterialButton, 'Switch to Page $num');

    expect(pageSwitcher, findsWidgets);
    await tap(pageSwitcher.first);
    await pumpAndSettle();
    await pageIsFound();
    var finder = find.widgetWithText(RouteTestPage, 'Page $num Page Body');
    expect(finder, findsOneWidget,
        reason:
            'Tapping the button switched routes to page $num, found: $allElements');
  }

  Future pageIsFound() async {
    expect(find.byType(RouteTestNotFoundPage), findsNothing,
        reason: "No NOT FOUND page");
  }

  Future openRecordPage(String recordId) async {
    var textFormField = find.byType(TextFormField);
    expect(textFormField, findsOneWidget);
    await enterText(textFormField, recordId);
    await tap(find.widgetWithText(MaterialButton, 'Load Record'));
    await pumpAndSettle();
    await pageIsFound();
    expect(find.widgetWithText(RouteTestPage, 'Record $recordId Page Body'),
        findsOneWidget,
        reason:
            'Entering text and tapping button opened Record Page $recordId');
  }
}

class RouteTestTab extends StatelessWidget {
  final String title;
  final WidgetListBuilder children;

  RouteTestTab.ofChildren({
    required this.title,
    required List<Widget> children,
  }) : children = ((context) => children);

  const RouteTestTab({required this.title, required this.children});

  @override
  Widget build(BuildContext context) {
    return Scaffold(
        appBar: AppBar(
          title: Text(title),
        ),
        body: Center(
          child: Column(
            children: [
              Text("${title} Page Body"),
              ...children(context),
            ],
          ),
        ));
  }
}

class RouteTestPage extends StatelessWidget {
  final String title;
  final WidgetListBuilder children;

  RouteTestPage.ofChildren({
    required this.title,
    required List<Widget> children,
  }) : children = ((context) => children);

  factory RouteTestPage.ofTabs(
      {Key? key,
      String? title,
      required int? thisPage,
      required int pageCount}) {
    var pageNums = [];
    for (var i = 1; i <= pageCount; ++i) {
      if (i != thisPage) {
        pageNums.add(i);
      }
    }
    return RouteTestPage(
        title: title ?? "Page $thisPage",
        children: (context) => [
              for (var targetPage in pageNums)
                if (targetPage != thisPage)
                  MaterialButton(
                    child: Text("Switch to Page ${targetPage}"),
                    onPressed: () {
                      Navigator.pushNamed(context, "/page$targetPage");
                    },
                  ),
            ]);
  }

  const RouteTestPage({required this.title, this.children = noChildren});

  @override
  Widget build(BuildContext context) {
    return Scaffold(
        appBar: AppBar(
          title: Text(title),
        ),
        body: Center(
          child: Column(
            children: [
              Text("${title} Page Body"),
              ...children(context),
            ],
          ),
        ));
  }
}

final NotFoundRoute = AppPageRoute(
  "/notFound",
  (context, _) => RouteTestNotFoundPage(),
  (_) => RouteParams.of(_),
);

class RouteTestNotFoundPage extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return RouteTestPage(title: "Not Found");
  }
}

List<Widget> noChildren(BuildContext context) => [];

class RouteTestTabs extends StatelessWidget {
  final int pageCount;
  final int? thisPage;
  const RouteTestTabs({Key? key, required this.pageCount, this.thisPage})
      : super(key: key);
  @override
  Widget build(BuildContext context) {
    var pageNums = [];
    for (var i = 1; i <= pageCount; i++) {
      if (i != thisPage) {
        pageNums.add(i);
      }
    }
    return SizedBox(
      height: 150,
      child: Column(
        children: [
          SizedBox(
              height: 75, child: Row(children: recordLoadWidgets(context))),
          SizedBox(
            height: 75,
            child: Row(
              children: [
                for (var targetPage in pageNums)
                  MaterialButton(
                    child: Text("Switch to Page ${targetPage}"),
                    onPressed: () {
                      Navigator.pushNamed(context, "/page${targetPage}");
                    },
                  ),
              ],
            ),
          ),
        ],
      ),
    );
  }
}

List<Widget> recordLoadWidgets(BuildContext context) {
  var inputValue = '';
  return [
    SizedBox(
      width: 100,
      child: TextFormField(
        initialValue: inputValue,
        onChanged: (v) => inputValue = v,
      ),
    ),
    MaterialButton(
      child: Text('Load Record'),
      onPressed: () {
        Navigator.pushNamed(context, '/page1/${inputValue}');
      },
    )
  ];
}
