import 'dart:async';

import 'package:firebase_auth/firebase_auth.dart' as fb;
import 'package:flutter/cupertino.dart' hide Notification;
import 'package:flutter/material.dart' hide Notification;
import 'package:sunny_dart/helpers/logging_mixin.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_essentials/sunny_essentials.dart';
import 'package:sunny_lifecycle/app_info.dart';
import 'package:sunny_lifecycle/base_extractors.dart';
import 'package:sunny_lifecycle/lifecycle_main.dart';
import 'package:sunny_lifecycle/push/push.dart';
import 'package:sunny_sdk_core/auth.dart';

import 'lifecycle_aware.dart';
import 'lifecycle_extensions.dart';

typedef PushMessageConsumer = FutureOr<bool> Function(PushSource source, PushMessageContent message);

class SourcedPushMessage {
  final PushMessageContent message;
  final PushSource source;

  const SourcedPushMessage(this.message, this.source);
}

/// Service that configures FCM messaging.  Any component can subscribe to inbound push notifications via [PushNotificationService.pushStream], but
/// a [PushHandlerService] is required for exclusive handling of an inbound message.  Basically, if you want to
/// display something.
class PushNotificationService with LoggingMixin, LifecycleAwareMixin {
  fb.User? _auth;
  String? _pendingToken;
  final AppInfo appInfo;
  final _pushStream = StreamController<PushMessageContent>.broadcast();
  final _internalConsumers = <String, PushMessageConsumer>{};

  final PushPlatform _platform;
  @override
  final Stream<AuthUserProfile> userStateStream;

  Stream<PushMessageContent> get pushStream => _pushStream.stream;

  final Future<bool> Function(BuildContext context, fb.User user, String token) sendDevicePushToken;
  final dynamic Function(String? envelopeType, String? body) convertMessage;

  PushNotificationService(
    Getter<BuildContext?> context,
    this.appInfo, {
    required this.userStateStream,
    required this.sendDevicePushToken,
    required this.convertMessage,
  }) : _platform = pushPlatform() {
    onLogout(() async {
      /// Reset the state by copying the current token to _pendingToken.  That way
      /// when the next user logs in, they can immediately register for notifications
      _pendingToken = _pushNotificationToken;
    });

    initialize(context, context()!.userChanges).whenComplete(() {
      _bindListeners(context);
    });
  }

  Future initialize(Getter<BuildContext?> context, Stream<AuthUserProfile?> userChanges) async {
    userChanges.listen((user) {
      if (user == null) {
        return;
      }
      if (_auth?.uid != user.fbUser?.uid) {
        _auth = user.fbUser;
        if (user.fbUser != null && _pendingToken != null && context() != null) {
          doSendPushToken(context()!, _pendingToken!);
          _pendingToken = null;
        } else {
          log.info("No pending token to send");
        }
      }
    }, cancelOnError: false).watch(this);

    _platform.pushTokens.auto(this, onItem: (token) {
      if (context() != null) {
        _withNewToken(context()!, token);
      }
    });

    await _platform.initialize(convertMessage);

    /// Don't await this one, because failed permissions will just cause a hang
    _platform.requestPermissions().then((value) => registerFcmToken(context));
  }

  @override
  Future doShutdown() async {
    await _pushStream.close();
  }

  String? _pushNotificationToken;

  String? get pushNotificationToken => _pushNotificationToken;

  void _bindListeners(Getter<BuildContext?> context, {bool configureBackground = true}) {
    try {
      _platform.requestPermissions().then((_) async {
        if (context() == null) return;
        log.info("Registering FCM handlers");
        await _platform.subscribe(context()!);
        _platform.messages.auto(this, onItem: (message) {
          if (context() == null) return;
          handleMessage(context()!, message);
        });
        log.info(" --- FCM handlers are registered!!");
      });
    } catch (e, stack) {
      if (configureBackground == true) {
        log.warning("Unable to configure background fcm: $e", e, stack);
        _bindListeners(context, configureBackground: false);
      } else {
        log.severe("Unable to configure fcm: $e", e, stack);
      }
    }
  }

  void registerListener(String key, PushMessageConsumer consumer) {
    _internalConsumers[key] = consumer;
  }

  void removeListener(String key) {
    _internalConsumers.remove(key);
  }

  Future _handleInternalListeners(PushSource source, PushMessageContent message) async {
    log.info("> PUSH with ${_internalConsumers.keys}");
    for (var entry in _internalConsumers.entries) {
      try {
        log.info(" -> Starting pushMessage handler: ${entry.key}");
        final res = await entry.value(source, message);
        log.info("    <- Done with ${entry.key}: Did work: $res");
      } catch (e, stack) {
        log.warning(" -> # Error processing ${entry.key} handler", e, stack);
      }
    }
  }

  Future handleMessage(
    BuildContext context,
    SourcedPushMessage source,
  ) async {
    try {
      /// Push to stream - this allows non-UI access to the messages
      if (!_pushStream.isClosed) {
        _pushStream.add(source.message);
      }
      _handleInternalListeners(source.source, source.message);

      // final result = this.handlers.handleNotification(context, pushMessage);
      // if (result == null) {
      //   log.warning("No handler for $pushMessage");
      // } else {
      //   await result;
      // }
    } catch (e, stack) {
      log.severe("Error handling inbound message: $e", e, stack);
    }
  }

  Future registerFcmToken(
    Getter<BuildContext?> context,
  ) async {
    final token = await _platform.getToken();
    log.info("Got FCM Token: $token");
    if (context() == null) return;
    return await _withNewToken(context()!, token);
  }

  Future _withNewToken(BuildContext context, String? token) async {
    if (this._pushNotificationToken != token) {
      this._pushNotificationToken = token;

      if (_auth != null) {
        doSendPushToken(context, token!);
      } else {
        log.info("Got token but no auth, set token to pending: ${token!.truncate(10)}");
        _pendingToken = token;
      }
    }
  }

  Future doSendPushToken(BuildContext context, String token) async {
    log.info("Sending device token: ${token.truncate(10)} to ${_auth!.displayName}");
    try {
      /// Why would this be commented out???
      await sendDevicePushToken.call(context, _auth!, token);
    } catch (e, stack) {
      log.severe("Error sending token: $e", e, stack);
    }
  }

  @override
  bool operator ==(Object other) => identical(this, other) || other is PushNotificationService;

  @override
  int get hashCode => 0;
}

Future<dynamic> backgroundMessageHandler(Map<String, dynamic> message) async {
  final themes = AppBootstrap.current().themesBuilder(Brightness.dark);
  print("Got background notification: $message");
  return CupertinoTheme(
    data: themes.cupertinoTheme,
    child: Theme(
      data: themes.darkTheme,
      child: PlatformListTile(
        title: Text("Push Notification!"),
        subtitle: Text("Another thing"),
      ),
    ),
  );
}

class PushMessageType {
  static const EVENTS = "EVENTS";
  static const CONTENT_BLOCKS = "CONTENT_BLOCKS";
  static const DESTINATION_URL = "DESTINATION_URL";
}

class PushEventsBody {
  final String? eventType;
  final Map<String, dynamic>? payload;

  PushEventsBody({this.eventType, this.payload});
}

class PushDestinationUrlBody {
  final String url;

  PushDestinationUrlBody(this.url);
}

enum PushSource { background, foreground, resume }

extension PushMessageExt on PushMessageContent {
  bool get isEvent {
    return envelopeType == PushMessageType.EVENTS;
  }

  bool isEventType(dynamic type) {
    if (type is Iterable) {
      final set = type.mapToString().toSet();
      return set.contains(this.pushEventsBody?.eventType);
    } else {
      return this.pushEventsBody?.eventType == "$type";
    }
  }

  bool get isDestinationUrl => envelopeType == PushMessageType.DESTINATION_URL;

  String? get destinationUrl {
    final body = this.body;
    if (isDestinationUrl) {
      assert(body is PushDestinationUrlBody);
      return (body as PushDestinationUrlBody).url;
    } else {
      return null;
    }
  }

  bool isEntityEvent([String? entityType]) {
    final b = pushEventsBody;
    if (b == null) return false;

    if (b.eventType == "ENTITY_UPDATED" || b.eventType == "ENTITY_DELETED" || b.eventType == "ENTITY_CREATED") {
      return entityType == null || entityType == b.payload!["entityType"];
    } else {
      return false;
    }
  }

  String? get entityId {
    final b = pushEventsBody;
    if (b == null) return null;

    if (b.eventType == "ENTITY_UPDATED" || b.eventType == "ENTITY_DELETED" || b.eventType == "ENTITY_CREATED") {
      return b.payload!["entityId"] as String?;
    } else {
      return null;
    }
  }

  String? get entityType {
    final b = pushEventsBody;
    if (b == null) return null;

    if (b.eventType == "ENTITY_UPDATED" || b.eventType == "ENTITY_DELETED" || b.eventType == "ENTITY_CREATED") {
      return b.payload!["entityType"] as String?;
    } else {
      return null;
    }
  }

  PushEventsBody? get pushEventsBody {
    return body is PushEventsBody ? body as PushEventsBody? : null;
  }
}

class PushMessageContent {
  String? recipient;
  String? userId;
  bool? nativePush;
  Object? body;
  Notification? notification;
  String? envelopeType;

  PushMessageContent({
    this.recipient,
    this.userId,
    this.nativePush,
    this.body,
    this.notification,
    this.envelopeType,
  });

  String toString() {
    return 'PushMessage[recipient=$recipient, userId=$userId, nativePush=$nativePush, body=$body, notification=$notification, envelopeType=$envelopeType, ]';
  }

  PushMessageContent.fromJson(Map<String, dynamic>? json) {
    if (json == null) return;
    recipient = json['recipient'] as String?;
    userId = json['userId'] as String?;
    nativePush = json['nativePush'] as bool?;
    body = json[body as String] as Map<String, dynamic>?;
    notification = Notification.fromJson(json['notification'] as Map<String, dynamic>?);
    envelopeType = json['envelopeType'] as String?;
  }

  Map<String, dynamic> toJson() {
    return {
      'recipient': recipient,
      'userId': userId,
      'nativePush': nativePush,
      'body': body,
      'notification': notification,
      'envelopeType': envelopeType
    };
  }

  static List<PushMessageContent> listFromJson(List<dynamic>? json) {
    json = json ?? const [];
    return [
      for (var value in json) PushMessageContent.fromJson(value as Map<String, dynamic>),
    ];
  }

  static Map<String, PushMessageContent> mapFromJson(Map<String, Map<String, dynamic>>? json) {
    if (json == null) return {};
    return <String, PushMessageContent>{
      for (var entry in json.entries) entry.key: PushMessageContent.fromJson(entry.value),
    };
  }
}

class Notification {
  String? title;
  String? subtitle;
  String? body;
  Uri? imageUrl;

  Notification({this.title, this.subtitle, this.body, this.imageUrl});

  @override
  String toString() {
    return 'Notification[title=$title, subtitle=$subtitle, body=$body, imageUrl=$imageUrl, ]';
  }

  Notification.fromJson(Map<String, dynamic>? json) {
    if (json == null) return;
    title = json['title'] as String?;
    subtitle = json['subtitle'] as String?;
    body = json['body'] as String?;
    imageUrl = extractUri(json['imageUrl']);
  }

  Map<String, dynamic> toJson() {
    return {'title': title, 'subtitle': subtitle, 'body': body, 'imageUrl': imageUrl};
  }

  static List<Notification> listFromJson(List<dynamic>? json) {
    json = json ?? const [];
    return [
      for (var value in json) Notification.fromJson(value as Map<String, dynamic>),
    ];
  }

  static Map<String, Notification> mapFromJson(Map<String, Map<String, dynamic>>? json) {
    if (json == null) return {};
    return <String, Notification>{
      for (var entry in json.entries) entry.key: Notification.fromJson(entry.value),
    };
  }
}
