import 'package:flutter/cupertino.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:logging/logging.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_fluro/sunny_fluro.dart';
import 'package:sunny_lifecycle/f_routing.dart';

class FRouterChildInfoProvider extends PlatformRouteInformationProvider {
  RouteInformation? _value;
  final String debugName;
  final RouteInformationProvider parent;
  final Logger log;

  FRouterChildInfoProvider({required this.debugName, required this.parent})
      : log = Logger(debugName);

  @override
  void routerReportsNewRouteInformation(RouteInformation routeInformation) {
    log.info("router reports: ${routeInformation.location}");
    if (routeInformation.location != null) {
      parent.routerReportsNewRouteInformation(routeInformation);
      _value = routeInformation;
      notifyListeners();
    }
  }

  @override
  RouteInformation? get value => _value;
}

abstract class FRouterRootProvider
    implements RouteInformationProvider, ChangeNotifier {
  factory FRouterRootProvider(
      {RouteInformation? initialRoute,
      required FRouteState state}) = FRouterRootProviderPlatform;

  void dispose();

  void childRouterReportsNewRouteInformation(RouteInformation routeInformation);

  void childRouterReportsPop(Route popped);
}

class FRouterRootProviderPlatform extends PlatformRouteInformationProvider
    implements FRouterRootProvider {
  final FRouteState state;
  FRouterRootProviderPlatform(
      {RouteInformation? initialRoute, required this.state})
      : super(initialRouteInformation: initialRoute);

  static final log = Logger("frouter");
  @override
  Future<bool> didPushRoute(String route) {
    log.info("platform:didPushRoute: $route");
    // Look backwards in history for a match, then pop routes to match.
    var matches = state.matchHistory(route);
    if (matches.isNotEmpty) {
      // Pop the route
      for (var match in matches) {
        if (match.currTab != null) {
          if (match.currTab != state.currTab) {
            log.info(
                "platform:didPushRouteInformation > move tab ${match.currTab}");
            state.currTab = match.currTab;
          } else {
            var tabNavState =
                state.tabState[match.currTab!].navigatorKey.currentState!;
            if (tabNavState.canPop() == true) {
              tabNavState.popUntil((route) {
                var isPath = route.path == route;
                if (!isPath) {
                  log.info(
                      "platform:didPushRouteInformation ^ popped ${match.route}");
                }
                return isPath;
              });
            } else {
              log.warning(
                  "platform:didPushRouteInformation ! unable to pop ${match.currTab}");
            }
          }
        } else {
          //pop root?
          log.warning("platform:didPushRouteInformation can't pop root (TBD)");
        }
      }

      return SynchronousFuture(true);
    } else {
      /// How do we know where to push this route?

      log.warning(() {
        var str = "";
        str +=
            'platform:didPushRouteInformation No found routes for ${route}\n';
        str += '  in ->\n';
        if (state.history.isEmpty) {
          str += '    - no history\n';
        } else {
          str +=
              '    - ${state.history.map((element) => element.route).join('\n    - ')}';
        }
        return str;
      });
    }
    return super.didPushRoute(route);
  }

  @override
  Future<bool> didPushRouteInformation(RouteInformation routeInformation) {
    log.info(
        "platform:didPushRouteInformation: ${routeInformation.location} state: ${routeInformation.state}");
    return super.didPushRouteInformation(routeInformation);
  }

  @override
  void dispose() {
    super.dispose();
  }

  @override
  void removeListener(VoidCallback listener) {
    super.removeListener(listener);
  }

  @override
  void addListener(VoidCallback listener) {
    super.addListener(listener);
  }

  @override
  RouteInformation? get value {
    return super.value;
  }

  @override
  void routerReportsNewRouteInformation(RouteInformation routeInformation) {
    log.info(
        "platform:didPushRouteInformation: ${routeInformation.location} ${routeInformation.runtimeType}");
    return super.routerReportsNewRouteInformation(routeInformation);
  }

  @override
  void childRouterReportsNewRouteInformation(
      RouteInformation routeInformation) {
    log.info("childDidPushRouteInformation: $routeInformation");
    if (routeInformation.location != null) {
      // SystemNavigator.routeInformationUpdated(
      //     location: routeInformation.location, state: routeInformation.state);
    }
  }

  @override
  void childRouterReportsPop(Route popped) {
    log.info("childPop: $popped");
    // SystemNavigator.pop(animated: true);
  }
}

// ignore:unused_element
class _FRouterRootProvider extends RouteInformationProvider
    with ChangeNotifier, LoggingMixin {
  RouteInformation? _value;

  _FRouterRootProvider({RouteInformation? initialRoute})
      : _value = initialRoute;

  @override
  void dispose() {
    // platform.removeListener(_onPlatformRouteChange);
    super.dispose();
  }

  String get loggerName => "rootProvider";

  @override
  void routerReportsNewRouteInformation(RouteInformation routeInformation) {
    log.info("router -> root: ${routeInformation.location}");

    if (routeInformation.location != null) {
      SystemNavigator.routeInformationUpdated(
          location: routeInformation.location!, state: routeInformation.state);
    }

    if (_value != routeInformation) {
      _value = routeInformation;
    }
  }

  @override
  RouteInformation? get value => _value;
}

extension RoutePathExt on Route {
  String? get path {
    final self = this.settings;
    if (self is PathRouteSettings) {
      return self.resolvedPathOrName;
    } else {
      return self.name;
    }
  }
}
