import 'package:equatable/equatable.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:sunny_core_widgets/core_ext.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_fluro/sunny_fluro.dart';
import 'package:sunny_lifecycle/routing/route_history.dart';
import 'package:sunny_lifecycle/sunny_lifecycle.dart';
import 'package:sunny_sdk_core/auth/auth_user_profile.dart';

class FRouteState extends ChangeNotifier
    with ThrottledChangeNotifier, EquatableMixin, LoggingMixin {
  FRouteState({AuthUserProfile? loggedIn, bool? isInitted, int? initialTab})
      : _loggedIn = null,
        _currTab = initialTab,
        _isInitted = false {
    this.loggedIn = loggedIn;
  }

  bool _isInitted;

  bool get isInitted => _isInitted;

  AuthUserProfile? _loggedIn;

  AuthUserProfile? get loggedIn => _loggedIn;

  UserDetails? get userDetails => _loggedIn?.profile;

  TabSet? _tabState;

  TabSet get tabState => _tabState ?? TabSet.empty();

  set tabState(TabSet value) => _tabState = value;

  TabState? get currentTabState {
    if (currTab == null) return null;
    return tabState[currTab];
  }

  final Map<int?, List<PendingRoute>> _tabRoutes = {};

  Map<int?, List<PendingRoute>> get tabRoutes => _tabRoutes;

  List<PendingRoute> get currentTabRoutes =>
      _tabRoutes.putIfAbsent(currTab, () => []);

  List<PendingRoute> get allRoutes => [
        ...currentTabRoutes,
        ..._rootRoutes,
      ];

  List<PendingRoute> _rootRoutes = [];

  Iterable<PendingRoute> get rootRoutes => _rootRoutes;

  final _history = RouteHistory<int>();
  Iterable<RouteHistoryItem<int>> get history => _history.items;

  set rootRoutes(Iterable<PendingRoute> rootRoutes) {
    final pendingList = rootRoutes.toList();
    if (_rootRoutes == pendingList) {
      return;
    }

    _rootRoutes = pendingList;
    notifyWithLimit();
  }

  set loggedIn(AuthUserProfile? loggedIn) {
    log.info(loggedIn == null
        ? "[loggedIn=null]"
        : "[loggedIn=${loggedIn.fbUser != null},profile=${loggedIn.profile != null}]");

    if (_loggedIn == loggedIn &&
        _isInitted == ((_isInitted == true) || loggedIn != null)) return;

    log.info(
        "Route state login change: ${loggedIn?.fbUser?.displayName ?? 'None'}");

    _isInitted = (_isInitted == true) || loggedIn != null;
    _loggedIn = loggedIn;
    notifyWithLimit();
  }

  /// The currently selected tab, or null if the user is not logged in.  This
  /// allows us to pass through navigation events.
  int? _currTab;

  int? get currTab => _currTab;
  int get currTabIndex => int.parse(currTab?.toString() ?? '0');

  set currTab(int? currTab) {
    if (currTab == _currTab) return;

    _currTab = currTab;
    notifyListeners();
  }

  @override
  String toString() {
    return joinString((str) {
      str += "FRouteState{isInit:${_isInitted}";
      if (_loggedIn?.fbUser?.email != null)
        str += "loggedIn:${_loggedIn?.fbUser?.email}";
      str += "currTab:${_currTab ?? 'not set'}";
    }, ';');
  }

  @override
  List<Object?> get props => [
        currTab,
        _isInitted,
        _loggedIn?.fbUser?.email,
        _rootRoutes,
      ];

  void popRootRoute() {
    _rootRoutes.removeLast();
    _history.pop();
    notifyWithLimit();
  }

  bool pushRootRoute(AppRouteMatch? match, {int? tab, bool isPage = true}) {
    if (match == null) return false;
    final route = PendingRoute(match, isPage: isPage);

    _rootRoutes.add(route);
    _history.push(route.toHistory(currTab: tab));
    notifyWithLimit();
    return true;
  }

  void pushAnonymousRoute(String route, {int? currTab}) {
    _history.push(RouteHistoryItem(route, currTab, isPage: false));
  }

  void popChildRoute(int? tab) {
    final tabs = _tabRoutes.putIfAbsent(tab, () => []);
    tabs.removeLast();
    _history.pop();
    notifyWithLimit();
  }

  void popAnonymousRoute() {
    _history.pop();
  }

  List<RouteHistoryItem<int>> matchHistory(String route) {
    return _history.match(route);
  }

  void pushHistoryItem(RouteHistoryItem<int> history) {
    /// this will mismatch!
    _history.push(history);
  }

  void pushCurrentTabRoute(PendingRoute pendingRoute) {
    currentTabRoutes.add(pendingRoute);
    _history.push(pendingRoute.toHistory(currTab: currTab));
  }
}

const appValueKey = ValueKey<String>('/app');
const rootValueKey = ValueKey<String>('/');
const splashValueKey = ValueKey<String>('/splash');
const loginValueKey = ValueKey<String>('/login');

final rootKeys = [
  rootValueKey,
  appValueKey,
  splashValueKey,
  loginValueKey,
].map((_) => _.value).toSet();

extension AppRouteMatchUriBuilderExt on AppRouteMatch {
  String? get resolvedUri {
    return this.route?.routeUri(parameters);
  }
}

class PendingRoute extends Equatable {
  final AppRouteMatch? match;
  final bool isPage;

  PendingRoute(this.match, {this.isPage = true});

  @override
  List<Object> get props => [route, isPage];

  String get route =>
      match!.resolvedUri ?? illegalState("Must provide route uri");
}

extension IterablePagesExt on Iterable<PendingRoute> {
  Iterable<PendingRoute> get pages => this.where((p) => p.isPage);
}

extension PendingRouteToHistory on PendingRoute {
  RouteHistoryItem<T> toHistory<T extends Object>({T? currTab}) =>
      RouteHistoryItem<T>(this.route, currTab, isPage: isPage);
}
