import 'dart:async';

import 'package:dartxx/dartxx.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:provider/provider.dart';
import 'package:sunny_core_widgets/core_ext.dart';
import 'package:sunny_core_widgets/snapshots.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_fluro/sunny_fluro.dart';
import 'package:sunny_lifecycle/app_context_events.dart';
import 'package:sunny_lifecycle/routing/route_history.dart';
import 'package:sunny_lifecycle/routing2/frouter_parser.dart';
import 'package:sunny_lifecycle/routing2/route_observer.dart';

import 'fade_animation_page.dart';
import 'frouter_state.dart';

abstract class BaseRouterDelegate<S extends FRouteMatch>
    extends RouterDelegate<S> {
  FRouterDelegate? get parent;
}

class FRouterDelegate extends BaseRouterDelegate<FRouteMatch>
    with
        PopNavigatorRouterDelegateMixin<FRouteMatch>,
        ChangeNotifier,
        LoggingMixin,
        ThrottledChangeNotifier {
  FRouterDelegate(this.state, this.routes,
      // this.routeProvider,
      {required this.path,
      required FRouterDelegate parent,
      required this.initialRoute,
      this.historyObservers = const [],
      required this.navigatorKey,
      required this.preferUpdates})
      : parent = parent {
    parent._children[this.path] = this;
    init();

    // init(userChanges);
    // _observer.addListener(_onRouterChanges);
  }

  FRouterDelegate.root(this.state, this.routes,
      // this.routeProvider,
      {this.historyObservers = const [],
      required this.initialRoute,
      required this.path,
      required this.preferUpdates,
      required this.navigatorKey})
      : parent = null {
    init();
    // _observer.addListener(_onRouterChanges);
  }

  final String path;
  final FRouter routes;
  final FRouteState state;

  // final Widget Function(UserDetails details) loggedInWidget;
  // final Widget Function() loggedOutWidget;
  // final Widget Function() splashPageWidget;
  final bool preferUpdates;

  /// The starting route for this delegate
  final AppRoute? initialRoute;
  final List<HistoryObserver> historyObservers;
  final _heroController = HeroController();

  FRouterDelegate get rootDelegate {
    FRouterDelegate delegate = this;
    while (delegate.hasParent) {
      delegate = delegate.parent!;
    }
    return delegate;
  }

  @override
  final GlobalKey<NavigatorState> navigatorKey;

  final Map<String, FRouterDelegate> _children = {};
  final FRouterDelegate? parent;

  bool get hasParent => parent != null;
  bool get isRoot => parent == null;

  @override
  final Limiter limiter = Debouncer(delay: 300.ms, atBegin: true);

  StreamSubscription? _sub;
  String? debugLabel;

  bool _hasSetInitialRoute = false;

  void init() {
    state.focusedChild = calculateInitialRoute();

    /// Listens to FRouteState state changes, notifies any router listeners,
    /// which will also trigger a rebuild so downstream widgets can adjust
    /// accordingly
    state.addListener(_onStateChange);
  }

  /// Called when state changes, basically, the change is propagated
  void _onStateChange() {
    /// Update

    if (_hasSetInitialRoute) notifyWithLimit();
  }

  String calculateInitialRoute() => initialRoute?.route ?? routes.paths.first;

  @override
  Future<void> setInitialRoutePath(FRouteMatch configuration) {
    var initialRouteMatch = routes.matchRoute(state.focusedChild);
    _currentActive = FRouteMatch.page(initialRouteMatch!, this);
    state.pushRoute(configuration.match,
        isPage: true, replace: false, notify: false);
    rootState.pushHistoryItem(configuration);
    this._hasSetInitialRoute = true;
    return SynchronousFuture(null);
  }

  /// Called by the [Router] when the [Router.routeInformationProvider] reports that a
  /// new route has been pushed to the application by the operating system.
  ///
  /// Consider using a [SynchronousFuture] if the result can be computed
  /// synchronously, so that the [Router] does not need to wait for the next
  /// microtask to schedule a build.
  @override
  Future<void> setNewRoutePath(FRouteMatch configuration) {
    if (configuration.matchedBy == this) {
      if (_currentActive != configuration) {
        final match = configuration.match;
        log.info("new route: ${match.resolvedUri}");
        // _checkLocationForTabChange(resolvedUri);
        state.pushRoute(match, isPage: true, replace: preferUpdates);
        rootState.pushHistoryItem(configuration);
        makeActive();
        _notifyIfReady();
      }
    } else {
      return configuration.matchedBy.setNewRoutePath(configuration);
    }
    return SynchronousFuture(null);
  }

  _notifyIfReady() {
    if (_hasSetInitialRoute) notifyListeners();
  }

  void push(AppRoute route, {RouteParams? args}) {
    final match =
        this.matchAppRoute(route, args ?? const DefaultRouteParams(const {}));
    this.setNewRoutePath(match);
  }

  void pushNamed(String route) {
    final p = matchRoute(RouteInformation(location: route));
    this.setNewRoutePath(p);
  }

  void activateChild(FRouterDelegate child) {
    this.setNewRoutePath(
        FRouteMatch.page(routes.matchRoute(child.path)!, this));
  }

  FRouteMatch matchPage(String route) {
    return FRouteMatch.page(routes.matchRoute(route)!, this);
  }

  void makeActive() {
    if (hasParent) {
      parent!.activateChild(this);
      parent!.makeActive();
    }
  }

  @override
  void dispose() {
    log.info("Disposed routerDelegate");
    // _observer.dispose();
    if (isActive) {
      _sub?.cancel();
    }

    if (hasParent) {
      parent!._children.remove(this.path);
    }
    for (var child in [..._children.values]) {
      child.dispose();
    }
    state.removeListener(_onStateChange);
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    var stackPages = state.viewStack.pages;
    var pages = [
      if (!_hasSetInitialRoute)
        FadeAnimationPage(
            child: Center(
              child: loadingIndicator,
            ),
            route: '/loading')
      else
        FadeAnimationPage(
          key: ValueKey(stackPages.first.match!.resolvedUri!),
          route: stackPages.first.match!.resolvedUri!,
          child: NavigatorContextWidget(
            child: stackPages.first.match!.route!
                .asPageRoute()!
                .handleAny(context, stackPages.first.match!.parameters),
          ),
        ),
      for (var p in state.viewStack.skip(1).pages)
        FadeAnimationPage(
          route: p.route,
          child: p.match!.route
              .asPageRoute()!
              .handleAny(context, p.match!.parameters),
          key: ValueKey("pending-${p.route}"),
        ),
    ];
    log.info(
        "routes.build[pages=${pages.map((e) => e.name ?? e.key).join(";")}");
    return Provider.value(
      value: SunnyRouter(this),
      builder: (context, value) => Navigator(
        key: navigatorKey,
        onPopPage: _handlePagePop,
        observers: [
          _heroController,
          FNavigatorObserver(this),
        ],
        onGenerateRoute: (settings) {
          final fmatch = this.matchRoute(RouteInformation(
              location: settings.name, state: settings.arguments));
          final match = fmatch.match;
          final creator = this
              .routes
              .routeFactory
              .generateAny(match.route!, null, null, null);
          return creator(match.route!.route,
              match.route!.paramConverter!(match.parameters));
        },
        reportsRouteUpdateToEngine: false,
        // restorationScopeId: _restorationId,
        pages: pages,
      ),
    );
  }

  @override
  Future<bool> popRoute() {
    log.info("Got pop route (delegating)");
    state.popRoute();
    rootState.popHistory();
    return super.popRoute();
  }

  reportPop(FRouteInformation information) {
    // does this get double-called with _handlePagePop?
    if (information.isExternal) {
      rootState.popHistory();
    }
  }

  reportPush(FRouteInformation information) {
    if (information.isExternal) {
      final matchy = matchRoute(information);
      rootState.pushHistoryItem(matchy);
    } else {
      /// push it?
    }
  }

  bool _handlePagePop(Route<dynamic> route, dynamic res) {
    try {
      log.info("Popped $route (with $res)");

      final bool result = route.didPop(res);
      if (state.viewStack.isNotEmpty) {
        log.info(
            "  ^pending Remove pending ${state.viewStack.last.route}; didPop=$result");
        state.popRoute();
        _updateLocation();
        return true;
      } else {
        log.info("  !pending Return false; didPop=$result");
        return false;
      }
    } catch (e, stack) {
      log.severe("Error popping: $e", e, stack);
      return false;
    }
  }

  PendingRoute? get currentRoute {
    if (_children[state.focusedChild] != null) {
      return _children[state.focusedChild]!.currentRoute;
    } else if (state.viewStack.isNotEmpty) {
      return state.viewStack.last;
    }
  }

  @override
  String toString() {
    return 'FRouterDelegate{state: $state}';
  }

  FRouteMatch? _currentActive;

  @override
  FRouteMatch? get currentConfiguration {
    return null;
  }

  void _updateLocation() {
    final curr = currentRoute;
    if (curr != null) {
      SystemNavigator.routeInformationUpdated(location: curr.route);
    }
  }

  FRouteState get rootState => rootDelegate.state;

  AppRouteMatch? fromRouteInformation(RouteInformation info) {
    var matchRoute = routes.matchRoute(info.location!);
    if (matchRoute?.resolvedUri == null) return null;
    return matchRoute;
  }

  FRouteMatch? tryMatchRoute(RouteInformation routeInformation,
      [Set? checked]) {
    checked ??= {};
    if (checked.contains(navigatorKey)) {
      return null;
    }
    checked.add(this.navigatorKey);
    var match = routes.matchRoute(routeInformation.location!);
    if (match != null) return FRouteMatch.page(match, this);
    for (var child in _children.values) {
      var childMatch = child.tryMatchRoute(routeInformation, checked);
      if (childMatch != null) {
        return childMatch;
      }
    }
    if (hasParent) {
      log.info(
          "passToParent: [route=${routeInformation.location}] ${parent!.navigatorKey}");
      return parent!.tryMatchRoute(routeInformation, checked);
    }
    return null;
  }

  FRouteMatch? tryMatchAppRoute(AppRoute route, RouteParams params,
      [Set? checked]) {
    checked ??= {};
    if (checked.contains(navigatorKey)) {
      return null;
    }
    checked.add(this.navigatorKey);
    final contains = routes.containsRoute(route);
    if (contains) {
      return FRouteMatch.page(AppRouteMatch(route, params), this);
    } else {
      for (var child in _children.values) {
        var childMatch = child.tryMatchAppRoute(route, params, checked);
        if (childMatch != null) {
          return childMatch;
        }
      }
      if (hasParent) {
        log.info(
            "passToParent: [route=${route.route}] ${parent!.navigatorKey}");
        return parent!.tryMatchAppRoute(route, params, checked);
      }
    }

    return null;
  }

  FRouteMatch matchAppRoute(AppRoute appRoute, RouteParams params) {
    return tryMatchAppRoute(appRoute, params) ??
        FRouteMatch.page(AppRouteMatch.ofMap(routes.notFoundRoute!, {}), this);
  }

  FRouteMatch matchRoute(RouteInformation routeInformation) {
    return tryMatchRoute(routeInformation) ??
        FRouteMatch.page(AppRouteMatch.ofMap(routes.notFoundRoute!, {}), this);
  }
}

class SunnyRouter {
  static SunnyRouter of(BuildContext context, {bool listen = false}) {
    return Provider.of<SunnyRouter>(context, listen: listen);
  }

  static FRouterDelegate delegateOf(BuildContext context,
      {bool listen = false}) {
    return Provider.of<SunnyRouter>(context, listen: listen).delegate;
  }

  static FRouterDelegate? tryDelegateOf(BuildContext context,
      {bool listen = false}) {
    try {
      return delegateOf(context);
    } on ProviderNotFoundException {
      return null;
    }
  }

  static void push<T, P extends RouteParams>(
      BuildContext context, AppRoute<T, P> route,
      {P? args}) async {
    delegateOf(context).push(route, args: args);
  }

  static void pushNamed(BuildContext context, String route) async {
    delegateOf(context).pushNamed(route);
  }

  final FRouterDelegate delegate;

  const SunnyRouter(this.delegate);
}

extension FRouterMatchExt on FRouter {
  bool containsRoute(AppRoute route, {bool calculate = false}) {
    return this.handlesRoute(route.route, calculate: calculate);
  }
}

extension BuildContextSunnyExt on BuildContext {
  void pushNamed(String route) => SunnyRouter.pushNamed(this, route);

  void push(AppRoute route, {RouteParams? args}) =>
      SunnyRouter.push(this, route, args: args);
}
