import 'package:flutter/cupertino.dart';
import 'package:logging/logging.dart';
import 'package:sunny_dart/helpers.dart';
import 'package:sunny_dart/helpers/logging_mixin.dart';
import 'package:sunny_fluro/sunny_fluro.dart';

import 'fade_animation_page.dart';
import 'frouter_delegate.dart';

abstract class FRouterNavigatorObserver extends NavigatorObserver
    with LoggingMixin {
  void handlePoppedRoute(FRouteInformation information);

  void handlePushedRoute(FRouteInformation information);

  @override
  void didPush(Route<dynamic> route, Route<dynamic>? previousRoute) {
    handlePushedRoute(_toRouteInfo(route,
        "Did push ${route.settings} (previous ${previousRoute?.settings})"));
  }

  @override
  void didPop(Route<dynamic> route, Route<dynamic>? previousRoute) {
    handlePoppedRoute(
        _toRouteInfo(route, "Overlay pop ${route}, previous $previousRoute"));
  }

  @override
  void didRemove(Route<dynamic> route, Route<dynamic>? previousRoute) {
    log.warning("We're not equipped for this: ");
    log.warning("  $route");
    log.warning("  $previousRoute");
    // report(previousRoute,
    //     "Did remove ${route?.settings} (previous ${previousRoute?.settings})");
  }

  @override
  void didReplace({Route<dynamic>? newRoute, Route<dynamic>? oldRoute}) {
    var src = "Did replace ${newRoute?.settings} (old ${oldRoute?.settings})";
    handlePoppedRoute(_toRouteInfo(oldRoute, src));
    handlePushedRoute(_toRouteInfo(newRoute, src));
  }

  FRouteInformation _toRouteInfo(Route? route, String source) {
    final settings = route?.settings;
    if (settings is PathRouteSettings && settings.resolvedPathOrName != null) {
      return FRouteInformation.path(settings);
    } else if (settings is FadeAnimationPage) {
      _log.fine(
          "${this.runtimeType}[overlay=${settings.route}]; reason $source");

      return FRouteInformation.page(settings);
    } else {
      _log.fine("${this.runtimeType}[overlay=anon] $source");
      return FRouteInformation.external(route, debugLabel: source);
    }
  }
}

class FRouteInformation implements RouteInformation {
  final String? name;
  final String? location;
  final bool isRecognized;
  final bool isExternal;
  final String? debugLabel;

  FRouteInformation.external(Route? route, {this.debugLabel})
      : location = "/${uuid()}",
        isExternal = true,
        isRecognized = false,
        name = route?.settings.name;

  FRouteInformation.path(PathRouteSettings page, {this.debugLabel})
      : name = page.name,
        location = page.location!,
        isRecognized = page.resolvedPathOrName != null,
        isExternal = true;

  FRouteInformation.page(Page page, {this.debugLabel})
      : location = page.name,
        isExternal = false,
        isRecognized = true,
        name = page.name;

  @override
  Object? get state => null;

  @override
  String toString() {
    return '$name: ${debugLabel ?? 'No debug label'}';
  }
}

//
class FNavigatorObserver extends FRouterNavigatorObserver {
  final FRouterDelegate delegate;

  FNavigatorObserver(this.delegate);

  @override
  void handlePoppedRoute(FRouteInformation information) =>
      delegate.reportPop(information);

  @override
  void handlePushedRoute(FRouteInformation information) =>
      delegate.reportPush(information);
}
//
// /// Passes events from a child router up to the root.  It's assumed that
// /// the root provider will not receive these events without this connector
// /// and that it should only report the url out to the navigator
// class TabRouteObserver extends FRouterNavigatorObserver {
//   final FRouterDelegate delegate;
//
//   /// The  index of the tab sending the events
//   final int tabIndex;
//
//   TabRouteObserver(this.delegate, {required this.tabIndex});
//
//   @override
//   void handlePoppedRoute(RouteInformation? information) =>
//       delegate.reportTabPop(information);
//
//   @override
//   void handlePushedRoute(RouteInformation? information) =>
//       delegate.reportTabPush(information);
// }

extension PathRouteSettingsHelperExt on PathRouteSettings {
  String? get resolvedPathOrName => resolvedPath ?? this.route;
}

final _log = Logger("routes");

extension RouteToInformation on Route {}
