import 'dart:async';

import 'package:flutter/widgets.dart' hide Container;
import 'package:provider/provider.dart';
import 'package:sunny_core_widgets/core_ext/context_resolvers.dart';
import 'package:sunny_dart/helpers/logging_mixin.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_essentials/provided.dart';
import 'package:sunny_kiwi/sunny_kiwi.dart';
import 'package:sunny_sdk_core/api/api_registry.dart';
import 'package:sunny_sdk_core/sunny_sdk_core.dart';
import 'package:sunny_service_stubs/services.dart';

import 'lifecycle_main.dart';

typedef InjectorInit = FutureOr Function(Container container);

class SunnyInjector with LoggingMixin {
  final Container _container;
  final InjectorInit preInit;
  final InjectorInit postInit;
  Future? _unloading;

  Future? get unloading => _unloading;

  // final TimeZone timeZone;

  SunnyInjector({
    Container? container,

    /// Runs before the main intitialization
    this.preInit = _noop,

    /// Runs after the main singleton initialization
    this.postInit = _noop,
  }) : _container = container ?? Container();

  void register<T>(T instance, {bool? isSilent}) {
    assert(instance is! Type);
    _container.registerInstance<T, T>(instance, isSilent: isSilent);
  }

  void registerFactory<T>(T factory(Container c), {bool? isSilent}) {
    _container.registerFactory<T, T>(factory, isSilent: isSilent);
  }

  void registerSingleton<T>(T factory(Container c), {bool? isSilent}) {
    _container.registerSingleton<T, T>(factory, isSilent: isSilent);
  }

  void registerAs<S, T extends S>(T instance, {bool? isSilent}) {
    assert(instance is! Type);
    _container.registerInstance<S, T>(instance, isSilent: isSilent);
  }

  void unregister<T>({bool? isSilent}) {
    _container.unregister<T>(isSilent: isSilent);
  }

  // void registerFactory<T>(T create(Container container), {bool isSilent}) {
  //   _container.registerFactory<T, T>(create, isSilent: isSilent);
  // }

  Future<SunnyInjector> initialize(
    LifecycleInit init, {
    required NamedIconContainer icons,
  }) async {
    try {
      // _container.registerInstance(navState);
      // _container.registerInstance(locale);
      _container.registerInstance<NamedIconContainer, NamedIconContainer>(icons,
          isSilent: true);
      _container.registerInstance(ApiRegistry(init.authState.userStateStream));
      _container.registerInstance(init.userPrefs);
      _container.registerInstance(init.helpService);
      _container.registerInstance(init.authState);
      _container.registerInstance(init.appInfo);
      _container.registerInstance(init.deviceInfo);
      _container.registerInstance(init.analytics);
      _container.registerInstance(init.routing);

      await preInit.call(_container);

      await _container.initializeEagerSingletons();
      log.info("Eager singletons initialized");
      await postInit.call(_container);
      // final _factService = await _container.resolve<FactService>();
      // _container.registerInstance<IFactService, FactService>(_factService);

      // await _container.instance<MSchemaService>().ready;
      return this;
    } catch (e, stack) {
      log.shout("Error initializing sunny context: $e", e, stack);
      rethrow;
    }
  }

  T? resolve<T>([String? name]) =>
      _container.resolve<T>(name: name, autoRegister: true);

  T? get<T>([String? name]) =>
      _container.resolve<T>(name: name, autoRegister: true);

  T? tryResolve<T>([String? name]) =>
      _container.tryInstance<T>(name: name, autoRegister: true)?.value as T?;

  Future reset() {
    return _container.clear();
  }

  Future? shutdown() async {
    if (_unloading != null) return _unloading;
    _unloading = _container.clear().then((_) {
      _unloading = null;
    });
    return _unloading;
  }

  static SunnyInjector of(BuildContext context) => Provided.get(context);
}

/// Resolves first against a kiwi context as well as the normal Provider.
/// Registrations are done against kiwi only.
class DualResolver extends ProviderBuildContextResolver {
  final SunnyInjector container;

  DualResolver(this.container);

  @override
  T resolve<T>(BuildContext? context) {
    try {
      final inst = container.tryResolve<T>();
      return inst ?? super.resolve<T>(context);
    } on ProviderNotFoundException {
      throw "$T instance could not be found in the ReliveIt context or by using Provider";
    }
  }

  @override
  Widget? registerAll(BuildContext context, List<Inst> inits,
      {Widget? child, Key? key}) {
    inits.forEach((element) {
      element.register(context, this.container);
    });
    var mapped = inits.map((e) {
      return e.bridgeFromContext(context, this.container);
    }).toList();
    var regAll = super.registerAll(context, mapped, child: child, key: key);
    return regAll;
  }
}

extension InstContainerExt<T extends Object> on Inst<T> {
  void register(BuildContext bc, SunnyInjector context) {
    void _reg<X extends Object>(Inst<X> inst) {
      if (this.isFactory) {
        context.registerSingleton<X>((container) => inst.factory!(bc),
            isSilent: this.skipIfRegistered);
      } else {
        context.register<X>(
          inst.instance!,
          isSilent: this.skipIfRegistered,
        );
      }
    }

    this.typed(_reg);
  }

  Inst bridgeFromContext(BuildContext bc, SunnyInjector context) {
    Inst<X> _bridge<X extends Object>(Inst<X> inst) {
      return Inst<X>.factory((_) => context.resolve<X>()!,
          skipIfRegistered: true);
    }

    return this.typed(_bridge);
  }
}

void _noop(_) {}
