import 'dart:async';

import 'package:shared_preferences/shared_preferences.dart';
import 'package:sunny_dart/helpers/logging_mixin.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_sdk_core/model/user_pref_key.dart';
import 'package:sunny_sdk_core/services/sunny.dart';

UserPreferencesService get userPrefs => sunny.get();

class UserPreferencesService with LoggingMixin {
  UserPreferencesService(this.keyPrefix,
      [this.getSharedPrefs = SharedPreferences.getInstance]);

  SharedPreferences? _prefs;
  final String keyPrefix;
  final Future<SharedPreferences> Function() getSharedPrefs;

  Future<SharedPreferences> get shared async {
    return _prefs ??= (await getSharedPrefs());
  }

  /// A list of known prefixes - any preference that starts with this prefix should be cleared
  /// out when a new account is created.
  ///
  /// Services can use [registerPrefix] to request this behavior
  final Set<String> _knownPrefixes = {};

  Future<Set<UserPrefKey>> get knownKeys async {
    return (await shared)
        .getKeys()
        .where((key) => key.startsWith(keyPrefix))
        .map((key) => UserPrefKey.fromQualifiedKey(keyPrefix, key))
        .toSet();
  }

  Future<String?> set(UserPrefKey key, String? value) {
    return shared.string(prefName(key), value);
  }

  Future setStrings(UserPrefKey key, Iterable<String> value) {
    return shared.stringList(prefName(key), [...value.whereNotNull()]);
  }

  Future setTrue(UserPrefKey key) {
    return shared.string(prefName(key), "true");
  }

  Future setFalse(UserPrefKey key) async {
    return shared.string(prefName(key), "false");
  }

  Future<DateTime> setTimestamp(UserPrefKey key) {
    return setDateTime(key, DateTime.now());
  }

  Future<DateTime> setDateTime(UserPrefKey key, DateTime dateTime) async {
    shared.string(prefName(key), dateTime.toIso8601String());
    return dateTime;
  }

  Future remove(UserPrefKey key) {
    return shared.remove(prefName(key));
  }

  Future<T?> get<T>(UserPrefKey key) {
    return shared.get(prefName(key));
  }

  Future<String?> getString(UserPrefKey key) => get<String>(key);

  Future<List<String>?> getStrings(UserPrefKey key) {
    return shared.stringList(prefName(key));
  }

  Future<bool> getBool(UserPrefKey key) async {
    return (await shared.get(prefName(key))) == "true";
  }

  Future<DateTime?> getDateTime(UserPrefKey key) async {
    final v = await shared.string(prefName(key));
    return dateTimeOf(v);
  }

  String prefName(UserPrefKey key) {
    return "$keyPrefix${key.name}";
  }

  Future resetPreferences() async {
    final removals = <Future>[];
    final keys = await this.knownKeys;
    for (final key in keys) {
      for (var prefix in _knownPrefixes) {
        if (key.startsWith(prefix)) {
          log.info("Removing preference: ${prefName(key)}");
          removals.add(remove(key));
          continue;
        }
      }
    }
    await Future.wait(removals);
  }

  void registerPrefix(String prefix) {
    if (prefix.isNotEmpty == true) _knownPrefixes.add(prefix);
  }

  Future<Map<UserPrefKey, T?>> findByPrefix<T>(String prefix) async {
    final keys = await this.knownKeys;
    return <UserPrefKey, T?>{
      for (final key in keys)
        if (key.startsWith(prefix)) key: await get<T>(key),
    };
  }
}

extension FutureSharedPreferences on Future<SharedPreferences> {
  Future<String?> string(String key, [String? data]) async {
    final shared = await this;
    if (data.isNotNullOrBlank) {
      await shared.setString(key, data!);
      return data;
    } else {
      return shared.get(key)?.toString();
    }
  }

  Future remove(String key) async {
    return (await this).remove(key);
  }

  Future<T?> get<T>(String key) async {
    final v = (await this).get(key);
    if (T == bool && v is String) {
      return (v == 'true') as T;
    } else if (T == String && v is! String) {
      return v?.toString() as T?;
    } else {
      return v as T?;
    }
  }

  Future<List<String>?> stringList(String key, [Iterable<String>? data]) async {
    final shared = await this;
    if (data.isNotNullOrEmpty) {
      await shared.setStringList(key, data!.toList());
      return data.toList();
    } else {
      return shared.getStringList(key);
    }
  }
}
